use std::sync::{
    Arc,
    atomic::{AtomicBool, Ordering},
};

use cpal::traits::{DeviceTrait, HostTrait, StreamTrait};
use protocol::{
    Error,
    protocol::{DEFAULT_SAMPLE_RATE, Signal, SignalType},
};
use rodio::{OutputStream, OutputStreamHandle, Sink, buffer::SamplesBuffer};
use tokio::sync::{broadcast, mpsc};

use crate::gui::State;

struct SpeakerSink {
    speaker_id: u8,
    sink: Sink,
}

impl SpeakerSink {
    fn new(speaker_id: u8, output_stream_handle: &OutputStreamHandle) -> Self {
        let sink = Sink::try_new(&output_stream_handle).unwrap();
        Self { speaker_id, sink }
    }
}

pub async fn record(
    mut record_control: mpsc::Receiver<State>,
    record_sender: Arc<broadcast::Sender<f32>>,
) -> Result<(), Error> {
    let host = cpal::default_host();
    let input_device = host
        .default_input_device()
        .ok_or(Error::Record("Input Device".to_string()))?;
    let config = input_device
        .default_input_config()
        .map_err(|inner| Error::Record(inner.to_string()))?
        .into();
    println!("Recorder Stream Config = {:#?}", config);

    let input = move |data: &[f32], _: &cpal::InputCallbackInfo| {
        for &sample in data {
            if record_sender.receiver_count() > 0 && sample != 0.0 {
                if let Err(err_val) = record_sender.send(sample) {
                    eprintln!("Error: Microphone Send | {}", err_val);
                }
            }
        }
    };

    let input_stream = input_device
        .build_input_stream(&config, input, voice_error, None)
        .map_err(|inner| Error::Record(inner.to_string()))?;

    input_stream
        .pause()
        .map_err(|inner| Error::Record(inner.to_string()))?;

    tokio::task::block_in_place(|| {
        loop {
            match record_control.blocking_recv() {
                Some(message) => match message {
                    State::Active => input_stream
                        .play()
                        .map_err(|inner| Error::Record(inner.to_string()))?,
                    State::Passive => input_stream
                        .pause()
                        .map_err(|inner| Error::Record(inner.to_string()))?,
                    State::Loading => {}
                },
                None => {
                    input_stream
                        .pause()
                        .map_err(|inner| Error::Record(inner.to_string()))?;
                    return Ok(());
                }
            }
        }
    })
}

async fn mixer(
    output_stream_handle: OutputStreamHandle,
    mut play_receiver: broadcast::Receiver<Signal>,
    play_pause: Arc<AtomicBool>,
) {
    let mut speaker_list = vec![];

    while let Ok(signal) = play_receiver.recv().await {
        match signal.get_signal_type() {
            SignalType::AudioDatum => {
                let data = if play_pause.load(std::sync::atomic::Ordering::Relaxed) {
                    [signal.get_audio_datum()]
                } else {
                    [0.0]
                };

                let source = SamplesBuffer::new(2, DEFAULT_SAMPLE_RATE, data);

                match speaker_list.binary_search_by(|speaker_sink: &SpeakerSink| {
                    speaker_sink.speaker_id.cmp(&signal.get_speaker_id())
                }) {
                    Ok(speaker_sink_index) => {
                        let speaker_sink = speaker_list.get(speaker_sink_index).expect("Never");

                        speaker_sink.sink.append(source);
                    }
                    Err(_) => {
                        let speaker_sink =
                            SpeakerSink::new(signal.get_speaker_id(), &output_stream_handle);

                        speaker_sink.sink.append(source);

                        speaker_list.push(speaker_sink);
                        speaker_list.sort_by(|x, y| x.speaker_id.cmp(&y.speaker_id));
                    }
                }
            }
            SignalType::SpeakerLeft => {
                speaker_list
                    .retain(|speaker_sink| speaker_sink.speaker_id != signal.get_speaker_id());
            }
        }
    }
}

pub async fn play(
    mut play_control: mpsc::Receiver<State>,
    play_receiver: broadcast::Receiver<Signal>,
) {
    let (_output_stream, output_stream_handle) = OutputStream::try_default().unwrap();
    let play_pause = Arc::new(AtomicBool::new(true));

    let mixer = tokio::spawn(mixer(
        output_stream_handle,
        play_receiver,
        play_pause.clone(),
    ));

    tokio::task::block_in_place(|| {
        loop {
            match play_control.blocking_recv() {
                Some(requested_state) => match requested_state {
                    State::Active => {
                        play_pause.store(true, Ordering::Relaxed);
                    }
                    State::Passive => {
                        play_pause.store(false, Ordering::Relaxed);
                    }
                    State::Loading => {}
                },
                None => {
                    mixer.abort();
                    return;
                }
            }
        }
    });

    mixer.abort();
}

fn voice_error(err_val: cpal::StreamError) {
    eprintln!("Error: Voice Error | {}", err_val);
}
